// $30 Sweep Antenna Analyser for 5.8GHz Band
// find out more: http://qczek.beyondrc.com/30-sweep-antenna-analyzer-for-5-8ghz-band/
#include <HardwareSerial.h>

//#define SA_DEBUG

#define SA_GET_SETTINGS 0x01
#define SA_GET_SETTINGS_V2 0x09
#define SA_SET_POWER 0x02
#define SA_SET_CHANNEL 0x03
#define SA_SET_FREQUENCY 0x04
#define SA_SET_MODE 0x05


#define SA_POWER_25MW 0
#define SA_POWER_200MW 1
#define SA_POWER_500MW 2
#define SA_POWER_800MW 3

#define SA_SERIAL_PORT_SPEED      4800
enum SMARTAUDIO_VERSION {
  NONE,
  SA_V1,
  SA_V2
};

static const uint8_t V1_power_lookup[] = {
  7,
  16,
  25,
  40
};

typedef struct {
  SMARTAUDIO_VERSION vtx_version;
  uint8_t channel;
  uint8_t powerLevel;
  uint8_t mode;
  uint16_t frequency;
  int8_t ok;
  
}unifyStr;


uint8_t crc8(const uint8_t *data, uint8_t len)
{
  uint8_t crc = 0;
  uint8_t currByte;

  for (int i = 0 ; i < len ; i++) 
  {
    currByte = data[i];
    crc ^= currByte;
    for (int i = 0; i < 8; i++) 
    {
      if ((crc & 0x80) != 0) 
      {
        crc = (byte)((crc << 1) ^ 0xd5);
      } else 
      {
        crc <<= 1;
      }
    }
  }
  return crc;
}

double getRFPowerVoltage(void)
{
  double retVal;
  uint32_t sum;
  sum = 0;
  // do overslampling
  for (int i = 0; i < 256; i++)
  {
    sum += analogRead(A0);
  }
  retVal = sum / 64; 
  // calculate output based on calibration done in excel
  retVal = (double)0.00122 * retVal + (double)0.0123;
  return retVal;
}

double voltTodBm(const double volt)
{
  double dBm;
  // calculate output based on calibration done in excel
  dBm = (double)22.4 * volt - (double)59.3;
  return dBm;
}

// based on https://github.com/NightHawk32/SmartAudio-testing/blob/master/tbs_test.ino
static void sa_tx_packet(const uint8_t cmd, const uint32_t value, const uint8_t vtx_version)
{
  uint8_t buff[10];
  uint8_t packetLength = 0;
  buff[0] = 0x00;
  buff[1] = 0xAA; //sync
  buff[2] = 0x55; //sync
  buff[3] = (cmd << 1) | 0x01; //cmd

  switch (cmd)
  {
    case SA_GET_SETTINGS:
      buff[4] = 0x00; //length
      buff[5] = crc8(&buff[1],4);
      buff[6] = 0x00;
      packetLength = 7;
      break;
    case SA_SET_POWER:
      buff[4] = 0x01; //length
      buff[5] = (vtx_version == SA_V1) ? V1_power_lookup[value] : value;
      buff[6] = crc8(&buff[1], 5);
      buff[7] = 0x00;
      packetLength = 8;
      break;
    case SA_SET_CHANNEL:
      buff[4] = 0x01; //length
      buff[5] = value;
      buff[6] = crc8(&buff[1], 5);
      buff[7] = 0x00;
      packetLength = 8;
      break;
    case SA_SET_FREQUENCY:
      buff[4] = 0x02;
      buff[5] = (value>>8); //high byte first
      buff[6] = value;
      buff[7] = crc8(&buff[1], 6);
      buff[8] = 0x00;
      packetLength = 9;
      break;
    case SA_SET_MODE: //supported for V2 only: UNIFY HV and newer
      if (vtx_version == SA_V2)
      {
      }
      break;
  }
  for(int i=0;i<packetLength; i++){
    Serial1.write(buff[i]);
  }
}

unifyStr sa_unpack_rx_packet(const uint8_t *buff, const uint8_t len )
{
  unifyStr retVal;
  memset(&retVal, 0, sizeof(retVal));

  //verify packet
  uint8_t packetStart=0;
  for(int i=0;i<len-3;i++){
    if(buff[i]==0xAA && buff[i+1]==0x55 && buff[i+3]<len){
      packetStart=i+2;
      uint8_t len=buff[i+3];
      uint8_t crcCalc=crc8(&buff[i+2],len+1);
      uint8_t crc=buff[i+3+len];

      if(crcCalc==crc){
        switch(buff[packetStart]){
          case SA_GET_SETTINGS: //fall-through
          case SA_GET_SETTINGS_V2:
            #ifdef SA_DEBUG
            Serial.println("SA_GET_SETTINGS");
            #endif
            retVal.vtx_version = (buff[packetStart]==SA_GET_SETTINGS) ? SA_V1 : SA_V2;
            packetStart+=2; //skip cmd and length
            retVal.channel = buff[packetStart++];
            retVal.powerLevel = buff[packetStart++];
            retVal.mode = buff[packetStart++];
            retVal.frequency = ((uint16_t)buff[packetStart++]<<8)|buff[packetStart++];
            retVal.ok = 1;
            break;
          case SA_SET_POWER:
            #ifdef SA_DEBUG
            Serial.println("SA_SET_POWER");
            #endif
            packetStart+=2;
            retVal.powerLevel = buff[packetStart++];  
            retVal.ok = 1;    
            break;
          case SA_SET_CHANNEL:
            #ifdef SA_DEBUG
            Serial.println("SA_SET_CHANNEL");
            #endif
            packetStart+=2;
            retVal.channel = buff[packetStart++];
            retVal.ok = 1;
            break;
          case SA_SET_FREQUENCY:
            #ifdef SA_DEBUG
            Serial.println("SA_SET_FREQUENCY");
            #endif
            //TBD: Pit mode Freq
            packetStart+=2;
            retVal.frequency = ((uint16_t)buff[packetStart++]<<8)|buff[packetStart++];
            retVal.ok = 1;
            break;
          case SA_SET_MODE:
            //SA V2 only!
            break;
        }
        return retVal;

      }else
      {
        //Serial.println("CRC mismatch");
        return retVal;
      }
    }
  }
  
}


void setup()
{
  unifyStr params_in;
  unifyStr params_out;
    
  Serial.begin(9600);
  while (!Serial);             // Leonardo: wait for serial monitor

  Serial.println("Start");
  Serial1.begin(SA_SERIAL_PORT_SPEED,SERIAL_8N2);
  UCSR1B &= ~(1<<TXEN1);

  // set min RF power level
 params_in.powerLevel = 0;
 params_out = sp_do_command(SA_SET_POWER, params_in);
 sp_print_unify( params_out, getRFPowerVoltage() );
 
}

void sp_print_unify( const unifyStr params,  const double voltage )
{
    Serial.print("Version:");
    Serial.print(params.vtx_version);
    Serial.print(", Channel:");
    Serial.print(params.channel);
    Serial.print(", PowerLevel:");
    Serial.print(params.powerLevel);
    Serial.print(", Mode:");
    Serial.print(params.mode);
    Serial.print(", Frequency:");
    Serial.print(params.frequency);
    Serial.print(", OK:");
    Serial.print(params.ok);
    Serial.print(", RFVol:");
    Serial.print(voltage);
    Serial.print(", dBm:");
    Serial.println(voltTodBm(voltage));

    
}

unifyStr sp_do_command(const uint8_t cmd, const unifyStr params )
{
   unifyStr retVal;
   uint8_t buff[25];
   uint8_t rx_len = 0;
   uint8_t zeroes = 0;

   Serial1.begin(SA_SERIAL_PORT_SPEED,SERIAL_8N2);
   switch(cmd)
   {
      case SA_GET_SETTINGS_V2:
        sa_tx_packet(SA_GET_SETTINGS_V2,0,SA_V2);
        break;  
      case SA_GET_SETTINGS:
        sa_tx_packet(SA_GET_SETTINGS,0,SA_V2);
        break;  
      case SA_SET_FREQUENCY:
        sa_tx_packet(SA_SET_FREQUENCY,params.frequency,SA_V2);
        break;                
      case SA_SET_CHANNEL:
         sa_tx_packet(SA_SET_CHANNEL,params.channel,SA_V2);
         break;
      case SA_SET_POWER:
         sa_tx_packet(SA_SET_POWER,params.powerLevel,SA_V2);
         break;   
         
    }
    Serial1.end();//clear buffer, otherwise sa_tx_packet is received
    Serial1.begin(SA_SERIAL_PORT_SPEED,SERIAL_8N2);
    UCSR1B &= ~(1<<TXEN1); //deactivate tx --> rx mode listening for response
    
    // wait for response from VTX
    delay(100);
    // skip bytes which are send by TX line
    while(Serial1.available())
    {
      buff[rx_len]=Serial1.read();
      if(buff[rx_len] == 0)
      {
        zeroes++;
      }
      rx_len++;
    }
  
    if(rx_len>6)
    {
      //because rx is low in idle 0 is received
      //when calculating crc of 0 we have a crc match, so
      //when all bytes are 0 we should avoid parsing the input data
      if(rx_len == zeroes)
      {
        // clean read buffer
        while(Serial1.available())
        {
          Serial1.read();
        }
      }else
      {
        retVal = sa_unpack_rx_packet(buff,rx_len);
      }
      zeroes=0;
      rx_len=0;   
    }

    return retVal;
  
}



uint8_t buff[25];
uint8_t rx_len = 0;
uint8_t zeroes = 0;
uint32_t curFreq = 5500;



void loop()
{
  unifyStr params_in;
  unifyStr params_out;
        
  int incomingByte = 0; 
  if (Serial.available() > 0) 
  {
    incomingByte = Serial.read();
    switch(incomingByte)
    {
      case 's':
        params_out = sp_do_command(SA_GET_SETTINGS, params_in);
        sp_print_unify( params_out,getRFPowerVoltage() );
        break;
      case '1':
        params_in.frequency = 5300;
        params_out = sp_do_command(SA_SET_FREQUENCY, params_in);
        delay(500);
        sp_print_unify( params_out, getRFPowerVoltage() );
        break;   
       case '2':
        params_in.frequency = 5800;
        params_out = sp_do_command(SA_SET_FREQUENCY, params_in);
        delay(500);
        sp_print_unify( params_out,getRFPowerVoltage() );
        break;  
       case 'z':
        params_in.powerLevel = 0;
        params_out = sp_do_command(SA_SET_POWER, params_in);
        sp_print_unify( params_out,getRFPowerVoltage() );
        break;   
       case 'x':    
        params_in.powerLevel = 1;
        params_out = sp_do_command(SA_SET_POWER, params_in);
        sp_print_unify( params_out,getRFPowerVoltage() );
        break;    
       case 'c':    
        params_in.powerLevel = 2;
        params_out = sp_do_command(SA_SET_POWER, params_in);
        sp_print_unify( params_out,getRFPowerVoltage() );
        break;    
       case 'v':    
        params_in.powerLevel = 3;
        params_out = sp_do_command(SA_SET_POWER, params_in);
        sp_print_unify( params_out,getRFPowerVoltage() );
        break;    
         
       case 'g':
         for(uint32_t i=5300; i<= 6000; i+=5)
         {
            params_in.frequency = i;
            params_out = sp_do_command(SA_SET_FREQUENCY, params_in);
            delay(1000);
            //sp_print_unify( params_out,getRFPowerVoltage() );
            double voltage = getRFPowerVoltage();
            Serial.print(params_in.frequency);
            Serial.print("\t");
            Serial.print(voltage,4);
            Serial.print("\t");
            Serial.println(voltTodBm(voltage) );
         }
         break;
    }
  }
}



